#include "precomp.h"

CEnumConnectionPoints::CEnumConnectionPoints()
    : m_lCursor(0), m_lRefCount(1)
{
}

CEnumConnectionPoints::~CEnumConnectionPoints()
{
    IConnectionPoint* pCP;

    while(m_list.Remove(0, &pCP))
        pCP->Release();
}

IFACEMETHODIMP
CEnumConnectionPoints::QueryInterface(REFIID riid, void** ppvObject)
{
    __try
    {
        *ppvObject = nullptr;

        if(IsEqualIID(riid, IID_IUnknown))
            *ppvObject = static_cast<IUnknown*>(this);
        else if(IsEqualIID(riid, IID_IEnumConnectionPoints))
            *ppvObject = static_cast<IEnumConnectionPoints*>(this);

        if(*ppvObject)
        {
            static_cast<IUnknown*>(*ppvObject)->AddRef();
            return(S_OK);
        }
    }
    __except((GetExceptionCode() == EXCEPTION_ACCESS_VIOLATION) ? EXCEPTION_EXECUTE_HANDLER : EXCEPTION_CONTINUE_SEARCH)
    {
        return(E_INVALIDARG);
    }

    return(E_NOINTERFACE);
}

IFACEMETHODIMP_(ULONG)
CEnumConnectionPoints::AddRef()
{
    return(InterlockedIncrement(&m_lRefCount));
}

IFACEMETHODIMP_(ULONG)
CEnumConnectionPoints::Release()
{
    const ULONG ulResult = InterlockedDecrement(&m_lRefCount);

    if(!ulResult)
        delete this;

    return(ulResult);
}

IFACEMETHODIMP
CEnumConnectionPoints::Next(ULONG cConnections, LPCONNECTIONPOINT* ppCP, ULONG* pcFetched)
{
    ULONG   lCursor;
    ULONG   ulCnt;
    BOOL    bl = IsBadWritePtrT(pcFetched);

    if(!bl)
        *pcFetched = 0;

#ifdef _DEBUG
    if(IsBadWritePtr(ppCP, (cConnections * sizeof(*ppCP))))
#else
    if(!ppCP)
#endif
        return(E_POINTER);

    ZeroMemory(ppCP, (cConnections * sizeof(*ppCP)));

    if(bl)
    {
        if((cConnections > 1) || (pcFetched))
            return(E_INVALIDARG);

        ulCnt = 0;
        pcFetched = &ulCnt;
        bl = FALSE;
    }

    while(*pcFetched < cConnections)
    {
        do
        {
            if(*ppCP)
            {
                (*ppCP)->Release();
                *ppCP = nullptr;
            }

#ifndef _MSC_VER
            lCursor = InterlockedCompareExchangeAcquire(&m_lCursor, SIZE_MAX, SIZE_MAX);
#else
            // Microsoft uses Acquire semantics on volatile variables.
            lCursor = m_lCursor;
#endif

            if(!m_list.Get(lCursor, ppCP))
            {
                bl = TRUE;
                break;
            }

            (*ppCP)->AddRef();

        } while(InterlockedCompareExchangeRelease(&m_lCursor, lCursor + 1, lCursor) != lCursor);

        if(bl)
            break;

        ++(*pcFetched);
        ++ppCP;
    }

    return((*pcFetched == cConnections) ? S_OK : S_FALSE);
}

IFACEMETHODIMP
CEnumConnectionPoints::Skip(ULONG cConnections)
{
    HRESULT hr;
    ULONG   lCurrentValue, lNewValue;

    do
    {
        hr = S_OK;

#ifndef _MSC_VER
        lCurrentValue = InterlockedCompareExchangeAcquire(&m_lCursor, SIZE_MAX, SIZE_MAX);
#else
        // Microsoft uses Acquire semantics on volatile variables.
        lCurrentValue = m_lCursor;
#endif

        lNewValue = lCurrentValue + cConnections;

#if UINTPTR_MAX <= ULONG_MAX
        if((lNewValue > m_list.Size()) || (lNewValue < lCurrentValue))
#else
        if(lNewValue > m_list.Size())
#endif
        {
            lNewValue = static_cast<ULONG>(m_list.Size());
            hr = S_FALSE;
        }
    } while(InterlockedCompareExchangeRelease(&m_lCursor, lNewValue, lCurrentValue) != lCurrentValue);

    return(hr);
}

IFACEMETHODIMP
CEnumConnectionPoints::Reset()
{
    InterlockedExchange(&m_lCursor, 0);
    return(S_OK);
}

IFACEMETHODIMP
CEnumConnectionPoints::Clone(IEnumConnectionPoints** ppEnum)
{
    CEnumConnectionPoints*  pEnum;
    HRESULT                 hr = S_OK;

    if(IsBadWritePtrT(ppEnum))
        return(E_POINTER);

    *ppEnum = nullptr;

    if(nullptr == (pEnum = new (std::nothrow) CEnumConnectionPoints))
        return(E_OUTOFMEMORY);

    m_list.Enum([&hr, pEnum](IConnectionPoint* pCP) -> bool
    {
        if(FAILED(hr = pEnum->Add(pCP)))
        {
            if(hr != E_OUTOFMEMORY)
                hr = E_UNEXPECTED;

            return(false);
        }
        return(true);
    });

    if(SUCCEEDED(hr))
    {
#ifndef _MSC_VER
        pEnum->m_lCursor = InterlockedCompareExchangeAcquire(&m_lCursor, SIZE_MAX, SIZE_MAX);
#else
        // Microsoft uses Acquire semantics on volatile variables.
        pEnum->m_lCursor = m_lCursor;
#endif

        hr = pEnum->QueryInterface(IID_IEnumConnectionPoints, reinterpret_cast<void**>(ppEnum));
    }

    pEnum->Release();
    return(hr);
}

__checkReturn HRESULT
CEnumConnectionPoints::Add(__in IUnknown* pCPt)
{
    IConnectionPoint*   pCP = nullptr;
    HRESULT             hr;

    if(!pCPt)
        return(E_INVALIDARG);

    if(FAILED(hr = pCPt->QueryInterface(IID_IConnectionPoint, reinterpret_cast<void**>(&pCP))))
        return(hr);

    if(!m_list.Add(pCP))
    {
        hr = HRESULT_FROM_WIN32(GetLastError());
        pCP->Release();
    }

    return(hr);
}
