#include "precomp.h"

HRESULT HresultFromHidErrorCode(NTSTATUS status)
{
    // Note: HRESULT_FROM_WIN32 is an inline function, hence you get C4640 when
    // compiling with /Wall. This is why __HRESULT_FROM_WIN32 has been used instead.
    static const struct
    {
        NTSTATUS    status;
        HRESULT     hr;
    } sc_map[] =
    {
        { HIDP_STATUS_SUCCESS,          S_OK                                            },
        { HIDP_STATUS_BUFFER_TOO_SMALL, __HRESULT_FROM_WIN32(ERROR_INSUFFICIENT_BUFFER) },
        { HIDP_STATUS_INTERNAL_ERROR,   __HRESULT_FROM_WIN32(ERROR_INTERNAL_ERROR)      },
        { HIDP_STATUS_NOT_IMPLEMENTED,  E_NOTIMPL                                       }
    };

    for(uintptr_t x = 0; x < _countof(sc_map); ++x)
        if(status == sc_map[x].status)
            return(sc_map[x].hr);

    return(E_FAIL);
}

HRESULT HresultFromErrno()
{
    // Note: HRESULT_FROM_WIN32 is an inline function, hence you get C4640 when
    // compiling with /Wall. This is why __HRESULT_FROM_WIN32 has been used instead.
    static const struct
    {
        int err;
        HRESULT hr;
    } sc_map[] =
    {
        { EACCES,   E_ACCESSDENIED                                  },
        { ENOMEM,   E_OUTOFMEMORY                                   },
        { EINVAL,   E_INVALIDARG                                    },
        { EBADF,    E_HANDLE                                        },
        { ENOENT,   __HRESULT_FROM_WIN32(ERROR_FILE_NOT_FOUND)      },
        { EMFILE,   __HRESULT_FROM_WIN32(ERROR_TOO_MANY_OPEN_FILES) },
        { ENOEXEC,  __HRESULT_FROM_WIN32(ERROR_BAD_FORMAT)          },
        { ENOSPC,   __HRESULT_FROM_WIN32(ERROR_DISK_FULL)           }
    };

    const int err = errno;

    for(uintptr_t x = 0; x < _countof(sc_map); ++x)
        if(sc_map[x].err == err)
            return(sc_map[x].hr);

    return(E_FAIL);
}

BOOL RectScreenToClient(_In_ HWND hWnd, _Inout_ RECT* prc)
{
    POINT pt[] = { {prc->left, prc->top }, {prc->right, prc->bottom} };
    BOOL ret = ScreenToClient(hWnd, &pt[0]);

    if(ret)
    {
        if(FALSE != (ret = ScreenToClient(hWnd, &pt[1])))
        {
            prc->left = pt[0].x;
            prc->top = pt[0].y;
            prc->right = pt[1].x;
            prc->bottom = pt[1].y;
        }
    }

    return(ret);
}

BOOL EnableDlgItem(HWND hDlg, int nIDDlgItem, BOOL blEnable)
{
    HWND hWnd = GetDlgItem(hDlg, nIDDlgItem);
    return((hWnd) ? EnableWindow(hWnd, blEnable) : FALSE);
}

int MinGetSystemMetricsForDpi(__in int nIndex1, __in int nIndex2, __in UINT dpi)
{
    const int val1 = MyGetSystemMetricsForDpi(nIndex1, dpi);
    const int val2 = MyGetSystemMetricsForDpi(nIndex2, dpi);
    return(std::min(val1, val2));
}

void ListView_ScaleColumnHeaders(HWND hWndLV, UINT dpiFrom, UINT dpiTo)
{
    int cnt = Header_GetItemCount(ListView_GetHeader(hWndLV));

    while(cnt-- > 0)
        ListView_SetColumnWidth(hWndLV, cnt, Scale(ListView_GetColumnWidth(hWndLV, cnt), dpiFrom, dpiTo));
}

HRESULT BSTRFromStringA(const char* pstr, BSTR* pbstr)
{
    const int nLen = MultiByteToWideChar(CP_ACP, MB_PRECOMPOSED | MB_ERR_INVALID_CHARS, pstr, -1, nullptr, 0);

    if(!nLen)
    {
        *pbstr = nullptr;
        return(HRESULT_FROM_WIN32(GetLastError()));
    }

    if(nullptr != (*pbstr = SysAllocStringLen(nullptr, static_cast<UINT>(nLen) - 1U)))
    {
        if(MultiByteToWideChar(CP_ACP, MB_PRECOMPOSED | MB_ERR_INVALID_CHARS, pstr, -1, *pbstr, nLen))
            return(S_OK);

        const HRESULT hr = HRESULT_FROM_WIN32(GetLastError());
        SysFreeString(*pbstr);
        *pbstr = nullptr;
        return(hr);
    }

    return(E_OUTOFMEMORY);
}

static WORD GetCharTypeA(const char* pch)
{
    WORD w = 0;
    GetStringTypeExA(LOCALE_INVARIANT, CT_CTYPE1, pch, 1, &w);
    return(w);
}

void TrimA(char* psz)
{
    PCSTR pFirst = psz;
    PSTR pLast;

    while(GetCharTypeA(pFirst) & C1_SPACE)
        pFirst = CharNextA(pFirst);

    for(pLast = CharPrevA(pFirst, pFirst + strlen(pFirst)); (GetCharTypeA(pLast) & C1_SPACE); pLast = CharPrevA(pFirst, pLast))
        *pLast = '\0';

    if(pFirst > psz)
        MoveMemory(psz, pFirst, static_cast<size_t>((CharNextA(pLast) - pFirst) + 1) * sizeof(*psz));
}

HRESULT GetTypeInfoOfChipIdEnumeration(ISmartICE04* pDevice, ITypeInfo** ppTypeInfo)
{
    ITypeInfo* pTypeInfo = nullptr;
    FUNCDESC* pDesc = nullptr;
    HRESULT hr;
    UINT idx = 0;

    if(FAILED(hr = pDevice->GetTypeInfo(0, CThreadLocale::getLCID(), &pTypeInfo)))
        return(hr);

    for(UINT cnt = 0; SUCCEEDED(hr = pTypeInfo->GetFuncDesc(cnt, &pDesc)); ++cnt)
    {
        if(pDesc->memid == DISPID_SMARTICE04_CHIPID)
        {
            if((pDesc->invkind == INVOKE_PROPERTYGET) && (pDesc->elemdescFunc.tdesc.vt == VT_USERDEFINED) && (pDesc->elemdescFunc.tdesc.hreftype))
            {
                if(SUCCEEDED(hr = pTypeInfo->GetRefTypeInfo(pDesc->elemdescFunc.tdesc.hreftype, ppTypeInfo)))
                    idx = cnt;
            }
        }

        pTypeInfo->ReleaseFuncDesc(pDesc);
        pDesc = nullptr;

        if(idx)
            break;
    }

    pTypeInfo->Release();
    return(hr);
}

static HRESULT WINAPI MyTaskDialogImpl(HWND hWndParent, HINSTANCE hInstance, PCWSTR pszWindowTitle, PCWSTR, PCWSTR pszContent, TASKDIALOG_COMMON_BUTTON_FLAGS dwCommonButtons, PCWSTR pszIcon, int *pnButton)
{
    MSGBOXPARAMSW params;
    int button;

    params.cbSize               = sizeof(params);
    params.hwndOwner            = hWndParent;
    params.hInstance            = hInstance;
    params.lpszText             = pszContent;
    params.lpszCaption          = pszWindowTitle;
    params.dwStyle              = 0;
    params.lpszIcon             = pszIcon;
    params.dwContextHelpId      = 0;
    params.lpfnMsgBoxCallback   = nullptr;
    params.dwLanguageId         = LANG_USER_DEFAULT;

    if(dwCommonButtons & TDCBF_CANCEL_BUTTON)
        params.dwStyle |= MB_OKCANCEL;

    if(dwCommonButtons & (TDCBF_YES_BUTTON | TDCBF_NO_BUTTON))
        params.dwStyle |= (MB_YESNO - params.dwStyle);
    else if(dwCommonButtons & TDCBF_RETRY_BUTTON)
        params.dwStyle = MB_RETRYCANCEL;

    if(pszIcon == TD_WARNING_ICON)
        params.dwStyle |= MB_ICONWARNING;
    else if(pszIcon == TD_ERROR_ICON)
        params.dwStyle |= MB_ICONERROR;
    else if(pszIcon == TD_INFORMATION_ICON)
        params.dwStyle |= MB_ICONINFORMATION;
    else if((pszIcon) && (pszIcon != TD_SHIELD_ICON))
        params.dwStyle |= MB_USERICON;

    if(!pnButton)
        pnButton = &button;

    if(0 == (*pnButton = MessageBoxIndirectW(&params)))
        return(HRESULT_FROM_WIN32(GetLastError()));

    return(S_OK);
}

typedef HRESULT (WINAPI * TASKDIALOG)(HWND hwndParent, HINSTANCE hInstance, PCWSTR pszWindowTitle, PCWSTR pszMainInstruction, PCWSTR pszContent, TASKDIALOG_COMMON_BUTTON_FLAGS dwCommonButtons, PCWSTR pszIcon, int *pnButton);
HRESULT MyTaskDialog(HWND hWndParent, HINSTANCE hInstance, PCWSTR pszWindowTitle, PCWSTR pszMainInstruction, PCWSTR pszContent, TASKDIALOG_COMMON_BUTTON_FLAGS dwCommonButtons, PCWSTR pszIcon, int *pnButton)
{
    static TASKDIALOG s_fnTaskDialog;

    if(s_fnTaskDialog)
    {
        CThreadDpiAwarenessContext ctx(DPI_AWARENESS_CONTEXT_PER_MONITOR_AWARE_V2);

        if(SUCCEEDED(s_fnTaskDialog(hWndParent, hInstance, pszWindowTitle, pszMainInstruction, pszContent, dwCommonButtons, pszIcon, pnButton)))
            return(S_OK);

        return(MyTaskDialogImpl(hWndParent, hInstance, pszWindowTitle, pszMainInstruction, pszContent, dwCommonButtons, pszIcon, pnButton));
    }

#pragma warning(suppress: 4191 6309 6387)
    if(nullptr == (s_fnTaskDialog = reinterpret_cast<TASKDIALOG>(GetProcAddress(GetModuleHandle(TEXT("comctl32.dll")), "TaskDialog"))))
        s_fnTaskDialog = MyTaskDialogImpl;

    return(MyTaskDialog(hWndParent, hInstance, pszWindowTitle, pszMainInstruction, pszContent, dwCommonButtons, pszIcon, pnButton));
}

HRESULT RedimAndLockArray(SAFEARRAY* psa, long lStartAddress, ULONG len)
{
    SAFEARRAYBOUND  sab, sabNew;
    HRESULT         hr;
    long            lBound;
    void*           pDst;

    if(FAILED(hr = SafeArrayGetLBound(psa, 1, &sab.lLbound)))
        return(hr);

    if(FAILED(hr = SafeArrayGetUBound(psa, 1, &lBound)))
        return(hr);

    if(0 != (sab.cElements = ((lBound >= 0) ? static_cast<ULONG>((lBound - sab.lLbound) + 1) : 0)))
    {
        sabNew.lLbound = std::min(sab.lLbound, lStartAddress);
        sabNew.cElements = std::max(static_cast<ULONG>(sab.lLbound) + sab.cElements, static_cast<ULONG>(lStartAddress) + len) - static_cast<ULONG>(sabNew.lLbound);
    }
    else
    {
        sabNew.lLbound = lStartAddress;
        sabNew.cElements = len;
    }

    if((sabNew.lLbound != sab.lLbound) || (sabNew.cElements != sab.cElements))
    {
        if(FAILED(hr = SafeArrayRedim(psa, &sabNew)))
            return(hr);

        if(FAILED(hr = SafeArrayLock(psa)))
            return(hr);

        if(sabNew.lLbound < sab.lLbound)
        {
            if(SUCCEEDED(hr = SafeArrayPtrOfIndex(psa, &sabNew.lLbound, &pDst)))
                MoveMemory(&reinterpret_cast<BYTE*>(pDst)[sab.lLbound - sabNew.lLbound], pDst, sab.cElements);
        }

        if(SUCCEEDED(hr))
        {
            lBound = static_cast<long>(static_cast<ULONG>(sab.lLbound) + sab.cElements);

            if((lBound < lStartAddress) && (sab.cElements))
            {
                if(SUCCEEDED(hr = SafeArrayPtrOfIndex(psa, &lBound, &pDst)))
                    FillMemory(pDst, static_cast<size_t>(lStartAddress - lBound), 0xFF);
            }
            else
            {
                lBound = static_cast<long>(static_cast<ULONG>(lStartAddress) + len);

                if(lBound < sab.lLbound)
                {
                    if(SUCCEEDED(hr = SafeArrayPtrOfIndex(psa, &lBound, &pDst)))
                        FillMemory(pDst, static_cast<size_t>(sab.lLbound - lBound), 0xFF);
                }
            }
        }

        if(FAILED(hr))
            SafeArrayUnlock(psa);

        return(hr);
    }

    return(SafeArrayLock(psa));
}

static LRESULT CALLBACK DpiChangeMonitorWndProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam, UINT_PTR uIdSubclass, DWORD_PTR dwRefData)
{
    const LRESULT ret = DefSubclassProc(hWnd, uMsg, wParam, lParam);

    switch(uMsg)
    {
        case WM_DPICHANGED_BEFOREPARENT:
        case WM_DPICHANGED_AFTERPARENT:
            SendMessage(reinterpret_cast<HWND>(uIdSubclass), static_cast<UINT>(dwRefData), uMsg, 0);
            break;
        case WM_DESTROY:
            RemoveWindowSubclass(hWnd, DpiChangeMonitorWndProc, uIdSubclass);
            break;
        default:
            break;
    }

    return(ret);
}

BOOL InstallDpiChangeMonitor(HWND hWndParent, UINT uMsg)
{
    HWND hWnd = GetWindow(hWndParent, GW_CHILD);

    if(!hWnd)
        return(FALSE);

    return(SetWindowSubclass(hWnd, DpiChangeMonitorWndProc, reinterpret_cast<UINT_PTR>(hWndParent), uMsg));
}

static UINT WINAPI GetDpiForWindowStub(HWND hWnd)
{
    UINT dpi = 0;
    HDC hdc = GetDC(hWnd);

    if(hdc)
    {
        const UINT tmp = dpi = static_cast<UINT>(GetDeviceCaps(hdc, LOGPIXELSX));
        dpi += static_cast<UINT>(GetDeviceCaps(hdc, LOGPIXELSY));

        if(tmp && (dpi > tmp))
            dpi /= 2;

        ReleaseDC(hWnd, hdc);
    }

    return((dpi) ? dpi : USER_DEFAULT_SCREEN_DPI);
}

static UINT WINAPI GetDpiForSystemStub()
{
    return(GetDpiForWindowStub(nullptr));
}

static int WINAPI GetSystemMetricsForDpiStub(int nIndex, UINT dpi)
{
    return((GetSystemMetrics(nIndex) * static_cast<int>(dpi)) / static_cast<int>(GetDpiForSystemStub()));
}

UINT MyGetDpiForWindow(HWND hWnd)
{
    typedef UINT (WINAPI * GETDPIFORWINDOW)(HWND hWnd);
    static GETDPIFORWINDOW s_pfn;

    if(!s_pfn)
    {
        HINSTANCE hUser = GetModuleHandle(TEXT("user32.dll"));

        if(hUser)
        {
#pragma warning(suppress: 4191)
            s_pfn = reinterpret_cast<GETDPIFORWINDOW>(GetProcAddress(hUser, "GetDpiForWindow"));
        }

        if(!s_pfn)
            s_pfn = GetDpiForWindowStub;
    }

    return(s_pfn(hWnd));
}

int MyGetSystemMetricsForDpi(int nIndex, UINT dpi)
{
    typedef int (WINAPI * GETSYSTEMMETRICSFORDPI)(INT nIndex, UINT dpi);
    static GETSYSTEMMETRICSFORDPI s_pfn;

    if(!s_pfn)
    {
        HINSTANCE hUser = GetModuleHandle(TEXT("user32.dll"));

        if(hUser)
        {
#pragma warning(suppress: 4191)
            s_pfn = reinterpret_cast<GETSYSTEMMETRICSFORDPI>(GetProcAddress(hUser, "GetSystemMetricsForDpi"));
        }

        if(!s_pfn)
            s_pfn = GetSystemMetricsForDpiStub;
    }

    return(s_pfn(nIndex, dpi));
}

BOOL CenterWindow(HWND hWnd, HWND hWndParent)
{
    RECT rc, rcParent;

    if(IsWindow(hWndParent))
    {
        GetWindowRect(hWndParent, &rcParent);
    }
    else if((GetWindowLongPtr(hWnd, GWL_STYLE) & WS_CHILD) && (nullptr != (hWndParent = GetParent(hWnd))))
    {
        GetClientRect(hWndParent, &rcParent);
    }
    else if(nullptr != (hWndParent = GetWindow(hWnd, GW_OWNER)))
    {
        GetWindowRect(hWndParent, &rcParent);
    }
    else
    {
        MONITORINFO mi = {};
        mi.cbSize = sizeof(mi);

        if(GetMonitorInfo(MonitorFromWindow(hWnd, MONITOR_DEFAULTTONEAREST), &mi))
        {
            rcParent = mi.rcWork;
        }
        else
        {
            SystemParametersInfo(SPI_GETWORKAREA, 0, &rcParent, 0);
        }
    }

    GetWindowRect(hWnd, &rc);
    const int x = (((rcParent.right - rcParent.left) - (rc.right - rc.left)) / 2) + rcParent.left;
    const int y = (((rcParent.bottom - rcParent.top) - (rc.bottom - rc.top)) / 2) + rcParent.top;
    return(SetWindowPos(hWnd, nullptr, x, y, 0, 0, SWP_NOOWNERZORDER | SWP_NOSIZE | SWP_NOZORDER));
}
